import os
import sys
import argparse
from pathlib import Path
import importlib
from contextlib import contextmanager
from collections import OrderedDict
from dynamicmethod import dynamicmethod
from qtpy import API_NAME, QtCore, QtGui, QtSvg

try:
    from subprocess import run
except (ImportError, Exception):
    from subprocess import Popen

    def run(*args, check=False, **kwargs):
        proc = Popen(*args, **kwargs)
        proc.wait()
        if check and proc.returncode != 0:
            raise RuntimeError('The program failed to run properly')
        return proc

from resource_man.__meta__ import version as __version__
from resource_man.importlib_interface import \
    READ_API, FILES_API, Traversable, contents, is_resource, read_binary, read_text, files, as_file
from resource_man.interface import \
    ResourceNotAvailable, Resource, ResourceManagerInterface, ResourceManager, \
    get_global_manager, set_global_manager, temp_manager, add_manager, remove_manager, \
    clear, register_resource, register, register_data, register_directory, unregister, has_resource, \
    get_resources, get_resource, get_binary, get_text, registered_datas, \
    MISSING


is_py2 = sys.version_info < (3, 0)

# Qt rcc compiler path
QT_RCC_BIN = files(API_NAME).joinpath('rcc.exe')
if not QT_RCC_BIN.exists():
    QT_RCC_BIN = files(API_NAME).joinpath(API_NAME.lower() + '-rcc.exe')
try:
    if 'PyQt' in API_NAME:
        # pyrcc5 binary
        QT_RCC_BIN = 'pyrcc' + ''.join(c for c in API_NAME if c.isdigit())
except (ValueError, Exception):
    pass


__all__ = [
    'QFile', 'QPixmap', 'QIcon', 'QSvgWidget', 'create_qrc', 'compile_qrc', 'create_compiled', 'load_resource',
    'compiled_py_to_qtpy',

    'READ_API', 'FILES_API', 'Traversable', 'contents', 'is_resource', 'read_binary', 'read_text', 'files', 'as_file',

    'ResourceNotAvailable', 'Resource', 'ResourceManagerInterface', 'ResourceManager',
    'get_global_manager', 'set_global_manager', 'temp_manager', 'add_manager', 'remove_manager',
    'clear', 'register_resource', 'register', 'register_data', 'register_directory', 'unregister',
    'has_resource', 'get_resources', 'get_resource', 'get_binary', 'get_text', 'registered_datas',
    'MISSING',
    '__version__'
    ]


class _ResourceExt(Resource):
    @property
    def qt_name(self):
        """Return the shortcut name used in the compiled qrc file.
        This is not a part of the standard api and should not be used.
        """
        prefix = getattr(getattr(self, 'manager', None), 'prefix', None)
        if prefix:
            return ':/{prefix}/{alias}'.format(prefix=prefix, alias=self.alias)
        return ':/' + self.alias

    orig_eq = Resource.orig_eq = Resource.__eq__  # NEED Resource.orig_eq to be set as well.

    def __eq__(self, other):
        if isinstance(other, str):
            return other == self.qt_name or \
                   other == self.alias or \
                   other.replace('\\', '/') == self.package_path
        return self.orig_eq(other)


# Override Resource methods and properties to work like ResourceExt
Resource.qt_name = _ResourceExt.qt_name
Resource.__eq__ = _ResourceExt.__eq__  # This will help ResourceManager.has_resource and get_resource


QtCore_QFile = QtCore.QFile
QtGui_QIcon = QtGui.QIcon
QtGui_QPixmap = QtGui.QPixmap


def get_file(name, return_bytes=True, extension=None):
    """Return the Qt file name or binary data from the file.

    Args:
        name (str/bytes/Traversable/Resource): Find the file from the name or resource object.
        return_bytes (bool)[True]: If name is Traversable or Resource try reading the bytes.
            If False return Traversable.
        extension (str)[None]: If given str check the extension before returning a valid value.

    Returns:
        name (str/bytes/Traversable): Prefer returning the string filename.
            If unable to find the filename and return_bytes is True return the file bytes.
            If unable to find the filename and return_bytes is False return the Traversable (Path) object.
            If not found or the filename does not have the proper extension return ''.
    """
    file = QFile(name)
    if extension is not None and file.extension() != extension:
        return None
    filename = file.fileName()
    if filename is not None:

        return filename
    elif return_bytes:
        return file.read_bytes()


class QFile(QtCore_QFile):
    def __new__(cls, name=None, parent=None):
        if isinstance(name, QtCore.QObject):
            parent = name
            name = None

        obj = QtCore_QFile.__new__(cls, 'None', parent)
        obj._byts = None
        obj._filename = None
        obj._resource = None
        obj.given_name = name

        return obj

    def __init__(self, name=None, parent=None):
        super().__init__()

        if isinstance(name, QFile):  # This class
            self._byts = getattr(self, '_byts', None)
            self._filename = getattr(self, '_filename', None)
            self._resource = getattr(self, '_resource', None)
            self.given_name = getattr(self, 'given_name', None)
        elif isinstance(name, (bytes, str, Resource, Traversable, Path)):
            self.setFileName(name)

        # name = self.given_name
        # byts = self._byts
        # filename = self._filename
        # resource = self._resource

    # ===== QFile methods =====
    @dynamicmethod
    def exists(self, name=None):
        if isinstance(self, QtCore_QFile):
            if self._byts:
                return True

            has_resource = self._resource and (QtCore_QFile.exists(self._resource.qt_name) or
                                               self._resource.is_resource())
            has_filename = self._filename and QtCore_QFile.exists(self._filename)
            return has_resource or has_filename or \
                   QtCore_QFile.exists(QtCore_QFile.fileName(self))  # super() will crash. Check the filename
        return type(self)(name).exists()

    def basename(self):
        """Return the resource basename."""
        try:
            if self._resource:
                return self._resource.name
            return os.path.basename(self._filename)
        except (TypeError, ValueError):
            return ''

    def extension(self):
        """Return the extension of the resource name or filename."""
        ext = os.path.splitext(self.basename())[-1]
        return ext

    def fileName(self):
        return self._filename

    def setFileName(self, name):
        self._filename = None
        self._byts = None
        self._resource = None
        if isinstance(name, bytes):
            self._byts = name
        elif isinstance(name, Resource):
            self._resource = name
            self._filename = str(self._resource)
            self._byts = self._resource.read_bytes()
        elif isinstance(name, Traversable):
            self._byts = name.read_bytes()
        else:
            # Try to create the icon from the Qt name
            str_name = str(name)

            # Try to create the icon from the registered resource name
            try:
                self._resource = get_resource(str_name)
                if self._filename is None:
                    if QtCore_QFile.exists(self._resource.alias):
                        self._filename = self._resource.alias
                    elif QtCore_QFile.exists(self._resource.qt_name):
                        self._filename = self._resource.qt_name

                    # Try reading the binary data
                    self._byts = self._resource.read_bytes()
            except (ResourceNotAvailable, TypeError, ValueError, OSError, ImportError, Exception):
                pass

            # Try to find the filename from the qt_name
            try:
                if self._filename is None:
                    qt_name = ':/' + str_name
                    if QtCore_QFile.exists(str_name):
                        self._filename = str_name
                    elif QtCore_QFile.exists(qt_name):
                        self._filename = qt_name
            except (ResourceNotAvailable, TypeError, ValueError, Exception):
                pass

        if self._filename is None:
            return super(QFile, self).setFileName('None')
        else:
            return super(QFile, self).setFileName(self._filename)

    def rename(self, newName):
        ret = super(QFile, self).rename(newName)
        if ret:
            self._filename = newName
        return ret

    # ===== QBuffer methods =====
    def data(self):
        if self.byts is not None:
            return QtCore.QByteArray(self.byts)
        return QtCore.QByteArray()

    def setData(self, data):
        self.byts = None
        if data is not None:
            self.byts = bytes(data)

    buffer = data
    setBuffer = setData

    # ===== Resource methods =====
    @property
    def package(self):
        """Return the resource package."""
        try:
            return self._resource.package
        except (AttributeError, Exception):
            return None

    @property
    def name(self):
        """Return the resource name."""
        try:
            return self._resource.name
        except (AttributeError, Exception):
            return None

    @property
    def package_path(self):
        """Return the package and name path of the resource."""
        try:
            return self._resource.package_path
        except (AttributeError, Exception):
            return ''

    @property
    def alias(self):
        """Return the alias name identifier."""
        try:
            return self._resource.alias
        except (AttributeError, Exception):
            return self.fileName()

    # Backwards compatibility support
    identifier = alias

    def is_resource(self):
        try:
            return self._resource.is_resource()
        except (AttributeError, Exception):
            return False

    def files(self):
        return self._resource.files()

    @contextmanager
    def as_file(self):
        with as_file(self.files()) as file:
            yield file

    def contents(self):
        return self._resource.contents()

    def read_bytes(self):
        """Return the read bytes."""
        if self._byts is None and self._resource is not None:
            self._byts = self._resource.read_bytes()
        elif self._byts is None and self._filename is not None:
            try:
                with open(self._filename, 'rb') as f:
                    self._byts = f.read()
            except OSError:
                if self._filename.startswith(':/'):
                    if self.open(self.ReadOnly | self.Text):
                        self._byts = bytes(self.readAll())
                        self.close()

        if self._byts is None:
            raise ResourceNotAvailable('Invalid resource!')
        return self._byts

    read_binary = read_bytes

    def read_text(self, encoding='utf-8', errors='strict'):
        return self.read_bytes().decode(encoding, errors)


class QPixmap(QtGui_QPixmap):
    """Special QPixmap that can load from resource_man values."""
    def __new__(cls, *args, **kwargs):
        return super(QPixmap, cls).__new__(cls)

    def __init__(self, *args, **kwargs):
        load_data = None
        if len(args) >= 1 and isinstance(args[0], (Resource, str, bytes, Traversable)):
            # Try to find filename, Qt File, or importlib.resources read resource bytes.
            file = QFile(args[0])
            args = ('',) + args[1:]
            if file.exists():
                load_data = file.read_bytes()

        super(QPixmap, self).__init__(*args, **kwargs)

        if isinstance(load_data, bytes):
            self.loadFromData(load_data)


class QIcon(QtGui_QIcon):
    """Special QIcon that can load from resource_man values."""
    def __new__(cls, *args, **kwargs):
        return super(QIcon, cls).__new__(cls)

    def __init__(self, *args, **kwargs):
        is_valid = False
        if len(args) >= 1:
            if isinstance(args[0], str) and QIcon.hasThemeIcon(args[0]):
                args = (QtGui_QIcon.fromTheme(args[0]),) + args[1:]
                is_valid = True
            elif isinstance(args[0], (Resource, str, bytes, Traversable)):
                # Try to find filename, Qt File, or importlib.resources read resource bytes.
                file = QFile(args[0])
                args = ('',) + args[1:]
                if file.exists():
                    pixmap = QtGui_QPixmap()
                    pixmap.loadFromData(file.read_bytes())
                    args = (pixmap, ) + args[1:]
                    is_valid = True

        super(QIcon, self).__init__(*args, **kwargs)
        self.is_valid = is_valid

    @dynamicmethod
    def fromTheme(self, name, fallback=None):
        icn = QIcon(name)
        if not icn.is_valid and fallback is not None:
            icn = QIcon(fallback)

        if isinstance(self, QIcon):
            self.swap(icn)
            return self
        return icn


class QSvgWidget(QtSvg.QSvgWidget):
    """QSvgWidget with resource_man support."""
    def __new__(cls, *args, **kwargs):
        return super(QSvgWidget, cls).__new__(cls)

    def __init__(self, *args, **kwargs):
        load_data = None
        if len(args) >= 1 and isinstance(args[0], (Resource, str, bytes, Traversable)):
            # Try to find filename, Qt File, or importlib.resources read resource bytes.
            file = QFile(args[0])
            args = ('',) + args[1:]
            if file.exists() and file.extension().lower() == '.svg':
                load_data = file.read_bytes()

        super(QSvgWidget, self).__init__(*args, **kwargs)

        if isinstance(load_data, bytes):
            self.load(load_data)


@contextmanager
def context_file(file):
    """Use as_file or just yield the given file."""
    if isinstance(file, Traversable):
        with as_file(file) as fname:
            yield fname
    else:
        yield file


def create_qrc(filename="resource_man_compiled_resources.qrc", prefix='', resource_manager=None, main_module=None):
    """Make a Qt RCC file.

    Note:
        This registers the resource with an alias shortcut identifier.

        .. code-block:: python

            >>> resource_man.register('check_lib.check_sub', 'edit-cut.png', alias='edit-cut')

        The alias can be used with QIcon(":/edit-cut")

    Example:
        .. code-block:: python

            >>> import resource_man.qt as resource_man
            >>>
            >>> resource_man.register('check_lib.check_sub', 'edit-cut.png', 'edit-cut')
            >>>
            >>> resource_man.create_qrc()  # app_resources.qrc created

    Args:
        filename (str)['resource_man_compiled_resources.qrc']: Filename to create the .qrc file with.
        prefix (str)['']: qresource prefix.
        resource_manager (ResourceManager)[None]: Resource manger to use. If None use default global ResourceManager.
        main_module (str)[None]: This module will be imported to register resources before creating the .qrc file.

    Returns:
        filename (str): Absolute path of the filename that was written.
    """
    if resource_manager is None:
        resource_manager = get_global_manager()

    # Import the main module that registers the resources
    if isinstance(main_module, (str, Path)):
        main_module = str(main_module)
        if '/' in main_module or '\\' in main_module or '.' in main_module:
            sys.path.append(os.path.dirname(main_module))
            main_module = os.path.splitext(os.path.basename(main_module))[0]
        importlib.import_module(main_module)

    # Get all managers according to their prefix
    managers = OrderedDict([(None, [])])
    prefix = prefix or getattr(resource_manager, 'prefix', None) or None
    try:
        managers[prefix].append(resource_manager)
    except (KeyError, AttributeError, Exception):
        managers[prefix] = [resource_manager]
    for man in resource_manager.managers:
        prefix = getattr(man, 'prefix', None) or None
        try:
            managers[prefix].append(man)
        except (KeyError, AttributeError, Exception):
            managers[prefix] = [man]

    # Create the QRC File
    text = ['<!DOCTYPE RCC><RCC version="1.0">']

    for prefix, mans in managers.items():
        if prefix:
            text.append('<qresource prefix="{}">'.format(prefix))
        else:
            text.append('<qresource>')

        for man in mans:
            for resource in man.get_resources():
                if resource.is_resource():
                    with resource.as_file() as rsc_file:
                        path = os.path.relpath(str(rsc_file))
                        text.append('\t<file alias="{}">{}</file>'.format(resource.alias, path))

        text.append('</qresource>')

    text.append('</RCC>')

    with open(filename, "w") as file:
        file.write("\n".join(text))

    return os.path.abspath(filename)


def compile_qrc(filename="resource_man_compiled_resources.qrc", output=None, rcc_bin=QT_RCC_BIN, use_qtpy=True):
    """Create the python resource file from the .qrc file.

    Args:
        filename (str)['resource_man_compiled_resources.qrc']: QRC resrouce file.
        output (str)[None]: Name of the new python file that should be created.
            Default "resource_man_compiled_resources.py" or "resource_man_compiled_resources.rcc".
        rcc_bin (Traversable/Path/str)[files(API_NAME).joinpath(rcc.exe)]: Path to the rcc binary.
        use_qtpy (bool)[True]: If True replace the Qt API with qtpy.

    Returns:
        filename (str): Python filename of the saved resource file.
    """
    if output is None:
        output = os.path.splitext(filename)[0] + ".py"  # or .py and "import app_resources.py"

    with context_file(rcc_bin) as rcc:
        rcc = str(rcc)
        args = [rcc, filename, '-o', output]
        is_py = not output.endswith('.rcc')
        is_pyqt = 'pyrcc' in rcc
        is_old_pyside = 'pyside-rcc' in rcc
        is_pyside = not is_pyqt and not is_old_pyside

        if is_pyside:  # Pyside2+
            # If .py file and PySide2+ you need "-g python" arguments
            if is_py:
                args.extend(['-g', 'python'])
            else:
                # .rcc binary file for pyside
                args.append('-binary')
        elif is_old_pyside:
            # .rcc binary file for PySide is not supported
            if not is_py:
                args[3] = output = os.path.splitext(output)[0] + '.py'  # PySide rcc does not support rcc binary

            # Need to identify the python version
            if is_py2:
                args.append('-py2')
            else:
                args.append('-py3')
        elif is_pyqt and not is_py:
            # .rcc binary file for PyQt is not supported
            args[3] = output = os.path.splitext(output)[0] + '.py'  # PyQT rcc does not support rcc binary

        success = run(args, stdout=sys.stdout, stderr=sys.stderr).returncode == 0
        if success and is_py and use_qtpy:
            compiled_py_to_qtpy(output)

        return output


def compiled_py_to_qtpy(filename):
    """Modify the compiled .py file to use qtpy instead of PySide2 or PyQt5."""
    # Convert resources.py import of PySide2 or PyQt5 to qtpy
    with open(filename, 'r+') as f:
        lines = f.readlines()
        f.seek(0, 0)
        f.truncate()  # File size is not different must truncate

        # Cannot use f.read().replace(API_NAME, 'qtpy') because we allow given rcc_bin
        for i, line in enumerate(lines):
            if ' import QtCore' in line:  # line is "from PySide2 import QtCore"
                lines[i] = 'from qtpy import QtCore\n'  # Does not use Windows \r\n. Otherwise could use os.linesep
                break

        f.writelines(lines)


def create_compiled(filename="resource_man_compiled_resources.qrc", prefix='', main_module=None, output=None,
                    rcc_bin=QT_RCC_BIN, use_qtpy=True):
    """Make a Qt RCC .qrc file, compile the Qt RCC file to a .py or .rcc file, and remove the unneeded QRC file.

    Note:
        This registers the resource with an alias shortcut identifier.

        .. code-block:: python

            >>> resource_man.register('check_lib.check_sub', 'edit-cut.png', 'edit-cut')

        The alias can be used with QIcon(":/edit-cut")

    Example:
        .. code-block:: python

            >>> import resource_man.qt as resource_man
            >>>
            >>> resource_man.register('check_lib.check_sub', 'edit-cut.png', 'edit-cut')
            >>>
            >>> resource_man.create_qrc()  # resource_man_compiled_resources.qrc created
            >>> resource_man.compile_qrc()  # resource_man_compiled_resources.py created
            >>> os.remove('resource_man_compiled_resources.qrc')

    Args:
        filename (str)['resource_man_compiled_resources.qrc']: Filename to create the .qrc file with.
        prefix (str)['']: qresource prefix.
        main_module (str)[None]: This module will be imported to register resources before creating the .qrc file.
        output (str)[None]: Name of the new python file that should be created.
            Default "resource_man_compiled_resources.py" or "resource_man_compiled_resources.rcc".
        rcc_bin (Traversable/Path/str)[files(API_NAME).joinpath(rcc.exe)]: Path to the rcc binary.
        use_qtpy (bool)[True]: If True replace the Qt API with qtpy.

    Returns:
        filename (str): Python filename of the saved resource file.
    """
    create_qrc(filename, prefix=prefix, main_module=main_module)
    out = compile_qrc(filename, output=output, rcc_bin=rcc_bin, use_qtpy=use_qtpy)
    os.remove(filename)
    return out


def load_resource(filename='resource_man_compiled_resources.py'):
    """Load a Qt resource file."""
    try:
        if filename == 'resource_man_compiled_resources.py':
            import resource_man_compiled_resources
            return True
    except (ImportError, Exception):
        pass

    try:
        # Try to use a resource file path
        with context_file(filename) as fname:
            fname = str(fname)
            if fname.endswith('.rcc'):
                return QtCore.QResource.registerResource(fname)
            else:
                directory = os.path.dirname(fname)
                dir_added = directory not in sys.path  # Could convert to := later. Want to support backwards compatible
                if dir_added:
                    sys.path.append(directory)
                try:
                    importlib.import_module(os.path.splitext(os.path.basename(fname))[0])
                finally:
                    if dir_added:
                        sys.path.remove(directory)
                return True
    except (ImportError, Exception):
        return False


if __name__ == '__main__':
    P = argparse.ArgumentParser('Run a Qt resource helper.')

    # ===== Sub commands =====
    SUBP = P.add_subparsers(help='Qt Resource commands.')

    # Create QRC
    CREATE_QRC = SUBP.add_parser('create', help='Create the .qrc file.')
    CREATE_QRC.set_defaults(func=create_qrc)
    # Main module to import
    CREATE_QRC.add_argument('main_module', type=str, help='Main module that is imported and registers resources.')
    CREATE_QRC.add_argument('--filename', '-f', type=str, default='resource_man_compiled_resources.qrc')
    CREATE_QRC.add_argument('--prefix', '-p', type=str, default='')

    # Compile QRC
    COMPILE_QRC = SUBP.add_parser('compile', help='Compile the .qrc file into an importable .py file of binaries.')
    COMPILE_QRC.set_defaults(func=compile_qrc)
    COMPILE_QRC.add_argument('--filename', '-f', type=str, default='resource_man_compiled_resources.qrc')
    COMPILE_QRC.add_argument('--output', '-o', type=str, default=None)
    COMPILE_QRC.add_argument('--rcc_bin', type=str, default=QT_RCC_BIN)
    COMPILE_QRC.add_argument('--use_qtpy', '-q', type=bool, default=True, help='Convert compiled resources to qtpy.')

    # Run both
    RUN_P = SUBP.add_parser('run', help='Create and compile the resources.')
    RUN_P.set_defaults(func=create_compiled)
    # Main module to import
    RUN_P.add_argument('main_module', type=str, help='Main module that is imported and registers resources.')
    RUN_P.add_argument('--filename', '-f', type=str, default='resource_man_compiled_resources.qrc')
    RUN_P.add_argument('--prefix', '-p', type=str, default='')
    RUN_P.add_argument('--output', '-o', type=str, default=None)
    RUN_P.add_argument('--rcc_bin', type=str, default=QT_RCC_BIN)
    RUN_P.add_argument('--use_qtpy', '-q', type=bool, default=True, help='Convert compiled resources to qtpy.')

    # ===== Parse given command line arguments into keyword arguments =====
    ARGS, REMAIN = P.parse_known_args()
    KWARGS = {name: getattr(ARGS, name, None) for name in dir(ARGS) if not name.startswith('_')}

    # ===== Run the subcommand =====
    FUNC = KWARGS.pop('func')
    FUNC(**KWARGS)



from .__meta__ import version as __version__

from .importlib_interface import \
    READ_API, FILES_API, Traversable, contents, is_resource, read_binary, read_text, files, as_file, \
    rsc_files, rsc_as_file, rsc_read_binary, rsc_read_text, rsc_contents, rsc_is_resource

from .interface import \
    ResourceNotAvailable, Resource, ResourceManagerInterface, ResourceManager, \
    get_global_manager, set_global_manager, temp_manager, add_manager, remove_manager, \
    clear, register_resource, register, register_data, register_directory, unregister, has_resource, \
    get_resources, get_resource, get_binary, get_text, registered_datas, \
    MISSING




import os
import sys
import contextlib
import inspect
from pathlib import Path

files = None
as_file = None
read_text = None
read_binary = None
contents = None
is_resource = None
Traversable = Path

READ_API = 'resource_man'
FILES_API = 'resource_man'

# Try reading imports
try:
    from importlib.resources import contents, is_resource, read_binary, read_text
    from importlib.abc import Traversable
    READ_API = 'importlib.resources'
except (ImportError, Exception):
    pass  # DO NOT USE importlib_resources! It does not work with PyInstaller. Throws Can't open orphan path error!
    # try:
    #     from importlib_resources import contents, is_resource, read_binary, read_text
    #     from importlib_resources.abc import Traversable
    #     READ_API = 'importlib_resources'
    # except (ImportError, Exception):
    #     pass

# Try files imports
try:
    from importlib.resources import files, as_file
    FILES_API = 'importlib.resources'
except (ImportError, Exception):
    pass  # DO NOT USE importlib_resources! It does not work with PyInstaller. Throws Can't open orphan path error!
    # try:
    #     from importlib_resources import files, as_file
    #     FILES_API = 'importlib_resources'
    # except (ImportError, Exception):
    #     pass


__all__ = [
    'READ_API', 'FILES_API', 'EXE_PATH',
    'Traversable', 'contents', 'is_resource', 'read_binary', 'read_text', 'files', 'as_file',
    'rsc_files', 'rsc_as_file', 'rsc_read_binary', 'rsc_read_text', 'rsc_contents', 'rsc_is_resource'
    ]


EXE_PATH = getattr(sys, '_MEIPASS', os.path.dirname(sys.executable))

if not hasattr(Traversable, 'read_bytes'):
    def path_read_bytes(self):
        with open(str(self), 'rb') as f:
            return f.read()

    Traversable.read_bytes = path_read_bytes

if not hasattr(Traversable, 'read_text'):
    def path_read_text(self, encoding='utf-8', errors='strict'):
        with open(str(self), 'r', encoding=encoding, errors=errors) as f:
            return f.read()

    Traversable.read_text = path_read_text


def rsc_files(module):
    if isinstance(module, str):
        # Check for the executable path
        if getattr(sys, 'frozen', False):
            path = Path(EXE_PATH) / module.replace('.', '/')
            if path.with_suffix('').name == '__init__':
                path = path.parent
            if path.exists():
                return path

        # Find the module path from the imported module
        if '.' in module:
            # Import the top level package and manually add a directory for each "."
            toplvl, remain = module.split('.', 1)
        else:
            toplvl, remain = module, ''

        # Get or import the module
        try:
            module = sys.modules[toplvl]
            path = Path(inspect.getfile(module))
        except (KeyError, Exception):
            try:
                module = __import__(toplvl)
                path = Path(inspect.getfile(module))
            except (ImportError, Exception):
                module = toplvl
                path = Path(module)

        # Get the path of the module
        if path.with_suffix('').name == '__init__':
            path = path.parent

        # Find the path from the top level module
        for pkg in remain.split('.'):
            path = path.joinpath(pkg)
    else:
        path = Path(inspect.getfile(module))
    if path.with_suffix('').name == '__init__':
        path = path.parent
    return path


@contextlib.contextmanager
def rsc_as_file(path):
    p = path

    # Find this path from the executable
    if (isinstance(p, str) and not os.path.exists(p)) or (isinstance(p, (Path, Traversable)) and not p.exists()):
        p = os.path.join(EXE_PATH, str(path))
    if (isinstance(p, str) and not os.path.exists(p)) or (isinstance(p, (Path, Traversable)) and not p.exists()):
        p = os.path.join(EXE_PATH, '', str(path))

    yield Path(p)  # Documentation says should be Path object, but I noticed it was a string.


def rsc_read_binary(package, resource):
    """Return the binary contents of the resource."""
    return files(package).joinpath(resource).read_bytes()


def rsc_read_text(package, resource, encoding='utf-8', errors='strict'):
    """Return the decoded string of the resource.

    The decoding-related arguments have the same semantics as those of
    bytes.decode().
    """
    return files(package).joinpath(resource).read_text(encoding, errors)


def rsc_contents(package):
    """Return an iterable of entries in 'package'.

    Note that not all entries are resources.  Specifically, directories are
    not considered resources.  Use `is_resource()` on each entry returned here
    to check if it is a resource or not.
    """
    return iter([p.name for p in files(package).iterdir()])


def rsc_is_resource(package, name):
    """True if 'name' is a resource inside 'package'.

    Directories are *not* resources.
    """
    pkg = files(package)
    path = pkg.joinpath(name)
    if path.exists() and not path.is_dir():
        return True

    try:
        package_contents = set(contents(package))
    except (NotADirectoryError, FileNotFoundError):
        return False
    if name not in package_contents:
        return False
    # Just because the given file_name lives as an entry in the package's
    # contents doesn't necessarily mean it's a resource.  Directories are not
    # resources, so let's try to find out if it's a directory or not.
    path = Path(pkg.__spec__.origin).parent / name
    return path.is_file()


if read_text is None:
    read_text = rsc_read_text
    read_binary = rsc_read_binary
    contents = rsc_contents
    is_resource = rsc_is_resource
    Traversable = Path

if files is None:
    files = rsc_files
    as_file = rsc_as_file



name = 'resource_man'
version = '2.1.12'
description = 'Python resource manager to work with the standard library for regular python code and executables.'
url = 'https://github.com/justengel/resource_man'
author = 'Justin Engel'
author_email = 'jtengel08@gmail.com'



import os
import sys
import copy
import atexit
import contextlib
from .importlib_interface import EXE_PATH, Traversable, contents, is_resource, read_binary, read_text, files, as_file

try:
    from dataclasses import MISSING
except (ImportError, Exception):
    MISSING = object()


__all__ = [
    'ResourceNotAvailable', 'Resource', 'ResourceManagerInterface', 'ResourceManager',
    'get_global_manager', 'set_global_manager', 'temp_manager', 'add_manager', 'remove_manager',
    'clear', 'register_resource', 'register', 'register_data', 'register_directory', 'unregister',
    'has_resource', 'get_resources', 'get_resource', 'get_binary', 'get_text', 'registered_datas',
    'MISSING'
    ]


class ResourceNotAvailable(Exception):
    pass


class Resource:
    def __init__(self, package, name, alias=MISSING, manager=None, data=None, **kwargs):
        """Initialize the resource object.

        Args:
            package (str): Package or module name where the resource can be found (EX: "mylib.mysubpkg")
            name (str): Name of the resource (EX: "myimg.png").
            alias (str)[MISSING]: Shortcut alias name identifer for the resource.
                ... (Ellipsis) will be the name with the extension (EX: "myimg.png")
                None will be the name without the extension (EX: "myimg")
            manager (ResourceManager)[None]: Manager that holds this object.
            data (bytes/object)[None]: Stored or read in data.
            **kwargs (dict): Dictionary of keyword arguments to set as resource attributes.
        """
        self.manager = manager
        self.raw_alias = alias
        self.package = package
        self.name = name
        self.data = data
        self._context = None
        self._context_obj = None

        # Set all keyword arguments as attributes
        for k, v in kwargs.items():
            try:
                setattr(self, k, v)
            except (AttributeError, TypeError, ValueError, Exception):
                pass

    @property
    def package_path(self):
        """Return the package path."""
        return '/'.join((self.package.replace('.', '/'), self.name))

    @property
    def alias(self):
        """Return the alias name identifier."""
        if self.raw_alias is MISSING:
            return self.package_path
        elif self.raw_alias is ...:
            return self.name
        elif self.raw_alias is None:
            return os.path.splitext(self.name)[0]
        return self.raw_alias

    @alias.setter
    def alias(self, value):
        self.raw_alias = value

    # Backwards compatibility support
    identifier = alias

    def is_resource(self):
        """Return if this resource is file based.

        This is mainly used for collecting data files. Resource's that were given data with fake packages and
        names will return False.
        """
        try:
            return is_resource(self.package, self.name)
        except (AttributeError, TypeError, ValueError, Exception):
            try:
                f = self.files()
                return f.exists() and not f.is_dir()
            except (AttributeError, TypeError, ValueError, Exception):
                return False

    def files(self):
        try:
            return files(self.package).joinpath(self.name)
        except (AttributeError, TypeError, ValueError, Exception):
            return Traversable(self.package_path)

    @contextlib.contextmanager
    def as_file(self):
        f = self.files()
        try:
            with as_file(f) as file:
                yield file
        except (ValueError, TypeError, FileNotFoundError, Exception) as err:
            package_path = str(self.package_path)
            if os.path.exists(package_path):
                yield package_path
            else:
                raise err

    def contents(self):
        try:
            if '/' in self.name or '\\' in self.name:
                return [f.name for f in self.files().parent.iterdir()]
        except (ValueError, TypeError, Exception):
            pass
        try:
            return contents(self.package)
        except (ValueError, TypeError, Exception):
            return []

    def read_bytes(self):
        if isinstance(self.data, bytes):
            return self.data
        elif isinstance(self.data, str):
            return self.data.encode('utf-8')

        error = None
        try:
            self.data = read_binary(self.package, self.name)
            return self.data
        except (AttributeError, TypeError, OSError, Exception) as err:
            error = err
            try:
                self.data = self.files().read_bytes()
                return self.data
            except (AttributeError, TypeError, OSError, Exception):
                pass
        raise ResourceNotAvailable(str(error))

    read_binary = read_bytes

    def read_text(self, encoding='utf-8', errors='strict'):
        if isinstance(self.data, str):
            return self.data
        elif isinstance(self.data, bytes):
            return self.data.decode(encoding, errors)

        error = None
        try:
            self.data = read_text(self.package, self.name, encoding, errors)
            return self.data
        except (AttributeError, TypeError, OSError, Exception) as err:
            error = err
            try:
                self.data = self.files().read_text(encoding, errors)
                return self.data
            except (AttributeError, TypeError, OSError, Exception):
                pass
        raise ResourceNotAvailable(str(error))

    def _enter_context(self):
        """Use "as_file" to enter the with context block for the life of the application in order to get the filepath.
        """
        self._context = self.as_file()
        self._context_obj = self._context.__enter__()
        try:
            self._context_obj = self._context_obj.resolve()  # Get proper path capitalization
        except (AttributeError, Exception):
            pass
        atexit.register(self._exit_context)

    def _exit_context(self):
        """Exit the context object that was used with "as_file"."""
        try:
            self._context.__exit__()
        except (ResourceNotAvailable, OSError, TypeError, ValueError, Exception):
            pass
        finally:
            self._context_obj = None
            self._context = None

    def __str__(self):
        """Return the full string file path. I'm not sure if this will always work. "as_file" should be used."""
        try:
            if self._context_obj:
                return str(self._context_obj)

            # Use "as_file" to enter the context and return the filepath
            self._enter_context()
            if self._context_obj:
                return str(self._context_obj)

            # Use files to return the path. Should I even do this?
            filename = self.files()
            try:
                filename = filename.resolve()
            except (AttributeError, OSError, Exception):
                pass
            return str(filename)
        except (ResourceNotAvailable, OSError, TypeError, ValueError, Exception):
            pass

        # Return the simple package path from the working directory.
        pp = orig_pp = str(self.package_path)
        if not os.path.exists(pp):
            pp = os.path.join(EXE_PATH, pp)
            if not os.path.exists(pp):
                pp = os.path.join(EXE_PATH, '', pp)
            if not os.path.exists(pp):
                pp = orig_pp
        return pp

    def __fspath__(self):
        """Return the file path. This is not recommended. You should be using 'as_file' or 'read_text'."""
        return self.__str__()

    def __eq__(self, other):
        if isinstance(other, str):
            return other == self.alias or \
                   other.replace('\\', '/') == self.package_path
        return super().__eq__(other)

    def __repr__(self):
        kwargs = {'cls': self.__class__.__name__, 'package': self.package, 'name': self.name, 'alias': self.alias}
        if self.raw_alias:
            return '{cls}(package={package}, name={name}, alias={alias})'.format(**kwargs)
        else:
            return '{cls}(package={package}, name={name})'.format(**kwargs)


class ResourceManagerInterface(object):
    RESOURCE_CLASS = Resource

    def __init__(self, *resources, **kwargs):
        if not hasattr(self, 'managers'):
            self.managers = []

        super(ResourceManagerInterface, self).__init__()
        self.init(*resources, **kwargs)

    def __iter__(self):
        raise NotImplementedError

    def __contains__(self, item):
        raise NotImplementedError

    def __getitem__(self, item):
        raise NotImplementedError

    def __setitem__(self, key, value):
        raise NotImplementedError

    def append(self, resource):
        if hasattr(resource, 'manager') and resource.manager is None:
            resource.manager = self
        raise NotImplementedError

    def pop(self, key, **kwargs):
        raise NotImplementedError

    def init(self, *resources, **kwargs):
        """Initialize"""
        self.managers = list(kwargs.pop('managers', getattr(self, 'managers', [])))

        # Set keyword arguments as attributes
        for k, v in kwargs.items():
            setattr(self, k, v)

        # Add resources
        for resource in resources:
            self.append(resource)

    def add_manager(self, man):
        """Add a manager to help with saving all managers."""
        self.managers.append(man)

    def remove_manager(self, man):
        """Remove a linked manager."""
        try:
            self.managers.remove(man)
        except (TypeError, ValueError, Exception):
            pass

    def register_resource(self, rsc, **kwargs):
        """Register a resource.

        Args:
            rsc (Resource): Resource object to register and access later.

        Returns:
            rsc (Resource): Resource object that was registered
        """
        self.append(rsc)
        return rsc

    def register(self, package, name, alias=MISSING, **kwargs):
        """Register a resource. You can optionally have an alias name identifier.
        When using the alias the last resource registered with the same alias will be used.

        Args:
            package (str): Package name ('check_lib.check_sub')
            name (str): Name of the resource ('edit-cut.png')
            alias (str)[MISSING]: Shortcut alias name identifer for the resource.
                ... (Ellipsis) will be the name with the extension (EX: "myimg.png")
                None will be the name without the extension (EX: "myimg")
            **kwargs (dict): Dictionary of keyword arguments to set as attributes to the resource.
        """
        if 'manager' not in kwargs:
            kwargs['manager'] = self
        rsc = self.RESOURCE_CLASS(package, name, alias=alias, **kwargs)
        return self.register_resource(rsc)

    def register_data(self, data, package, name, alias=MISSING, **kwargs):
        """Register a plain data resource.

        Args:
            data (bytes/str/object)[None]: Plain data to store.
            package (str)[None]: Package name ('check_lib.check_sub')
            name (str)[None]: Name of the resource ('edit-cut.png')
            alias (str)[MISSING]: Shortcut alias name identifer for the resource.
                ... (Ellipsis) will be the name with the extension (EX: "myimg.png")
                None will be the name without the extension (EX: "myimg")
            **kwargs (dict): Dictionary of keyword arguments to set as attributes to the resource.
        """
        if 'manager' not in kwargs:
            kwargs['manager'] = self
        rsc = self.RESOURCE_CLASS(package=package, name=name, alias=alias, data=data, **kwargs)
        return self.register_resource(rsc)

    def register_directory(self, package, directory='', recursive=False, extensions=None, exclude=None, **kwargs):
        """Register all items in a directory.

        Note:
            This does not register the __init__.py module.

        Args:
            package (str): Package name ('check_lib.check_sub')
            directory (str)['']: Additional directory path.
            recursive (bool)[False]: If True iterate through subdirectories and find all files.
            extensions (list/str)[None]: List of extensions to register (".csv", ".txt", "" for "LICENSE" with no ext).
                If None register all.
            exclude (list/str)[None]: List of filenames to exclude.
            **kwargs (dict): Dictionary of keyword arguments to set as attributes to the resource.

        Returns:
            directory (list): List of Resource objects that were registered.
        """
        if isinstance(extensions, str):
            extensions = [extensions]
        if exclude is None:
            exclude = []
        elif isinstance(exclude, str):
            exclude = [exclude]
        directory = directory or ''

        folder = []
        pkg = files(package).joinpath(directory)
        if directory or recursive:
            if recursive:
                iter_dir = pkg.glob('**/*')
            else:
                iter_dir = pkg.iterdir()
            for f in iter_dir:
                name = str(f.relative_to(pkg))
                ext = os.path.splitext(name)[-1]
                if not f.is_dir() and (name not in exclude) and (extensions is None or ext in extensions):
                    path = os.path.join(directory, name).replace('\\', '/')  # Normalize path with the directory
                    folder.append(self.register(package, path, **kwargs))
        else:
            for name in contents(package):
                name = str(name)
                ext = os.path.splitext(name)[-1]
                f = pkg.joinpath(name)
                if not f.is_dir() and (name not in exclude) and (extensions is None or ext in extensions):
                    folder.append(self.register(package, name, **kwargs))

        return folder

    def unregister(self, rsc=None, name=None, alias=None):
        """Unregister a resource.

        Args:
            rsc (str/Resource)[None]: Resource to unregister or package name (with name argument) or alias.
            name (str)[None]: Required if package name given
            alias (str)[None]: Alias name identifier to find. This may also be given as first argument.

        Returns:
            rsc (Resource): Resource that was found

        Raises:
            error (ResourceNotAvailable): Resource was not found!
        """
        if isinstance(rsc, str) and isinstance(name, str):
            rsc = '/'.join((rsc.replace('.', '/'), name))
        elif isinstance(alias, str):
            rsc = alias

        for i in reversed(range(len(self))):
            resource = self[i]
            if resource == rsc:
                return self.pop(i)

        raise ResourceNotAvailable('Resource not found! If "package" given then the "name" argument is required.')

    def has_resource(self, rsc=None, name=None, alias=None):
        """Return if the registered resource exists (this can be the alias, resource, or package_path."""
        if isinstance(rsc, str) and isinstance(name, str):
            rsc = '/'.join((rsc.replace('.', '/'), name))
        elif isinstance(alias, str):
            rsc = alias

        return rsc in self

    def get_resources(self):
        """Return a list of registered resources.

        If multiple resources have the same alias the last one will be used.
        """
        return copy.copy(self)

    def get_resource(self, rsc, fallback=None, default=MISSING):
        """Return the found Resource object from the given resource, fallback, or default value.

        Args:
            rsc (str/Resource): Alias name identifier, package path, or Resource object that has a registered resource
            fallback (str)[None]: Fallback alias if the given alias was not registered.
            default (object)[MISSING]: Default value to return if the alias and fallback were not found.

        Returns:
            rsc (Resource): The found Resource object.
        """
        # Try finding the identifier
        try:
            return self[rsc]
        except (KeyError, ResourceNotAvailable):
            pass

        # Try fallback
        try:
            if isinstance(fallback, Resource):
                return fallback
            return self[fallback]
        except (KeyError, ResourceNotAvailable):
            pass

        # Check default
        if default is MISSING:
            raise ResourceNotAvailable('Resource "{}" not found'.format(rsc))
        return default

    def get_binary(self, rsc, fallback=None, default=MISSING):
        """Return the binary data for the Resource object from the given identifier or default.

        Args:
            rsc (str/Resource): Alias name identifier, package path, or Resource object that has a registered resource
            fallback (str)[None]: Fallback alias if the given alias was not registered.
            default (object)[MISSING]: Default value to return if the alias and fallback were not found.

        Returns:
            data (bytes): Binary data that was read.
        """
        rsc = self.get_resource(rsc, fallback, default)
        if isinstance(rsc, Resource):
            return rsc.read_binary()
        return rsc

    def get_text(self, rsc, fallback=None, default=MISSING, encoding='utf-8', errors='strict'):
        """Return the text data for the Resource object from the given identifier or default.

        Args:
            rsc (str/Resource): Alias name identifier, package path, or Resource object that has a registered resource
            fallback (str)[None]: Fallback alias if the given alias was not registered.
            default (object)[MISSING]: Default value to return if the alias and fallback were not found.
            encoding (str)['utf-8']: Encoding to convert binary data to text.
            errors (str)['strict']: Error handling code when converting the binary data to text.

        Returns:
            text (str): Text data that was read.
        """
        rsc = self.get_resource(rsc, fallback, default)
        if isinstance(rsc, Resource):
            return rsc.read_text(encoding=encoding, errors=errors)
        return rsc


class ResourceManager(list, ResourceManagerInterface):

    RESOURCE_CLASS = Resource

    def __init__(self, *resources, **kwargs):
        list.__init__(self)
        ResourceManagerInterface.__init__(self, *resources, **kwargs)

    __iter__ = list.__iter__
    __contains__ = list.__contains__

    def __getitem__(self, item):
        if isinstance(item, int):
            return list.__getitem__(self, item)

        # Search through all linked managers
        for man in [self] + self.managers:
            for rsc in reversed(man):
                if rsc == item:
                    return rsc

        raise ResourceNotAvailable("The requested resource was not found!")

    def __setitem__(self, key, value):
        if isinstance(key, int):
            list.__setitem__(self, key, value)
            return
        elif isinstance(key, self.RESOURCE_CLASS):
            key = key.alias or key.package_path

        # Find the resource and replace the identifier
        for i in reversed(range(len(self))):
            rsc = self[i]
            if rsc.alias == key:
                list.__setitem__(i, value)
                return

        # If not found add the resource to the list.
        self.append(value)

    def append(self, resource):
        if hasattr(resource, 'manager') and resource.manager is None:
            resource.manager = self
        return list.append(self, resource)

    pop = list.pop


RESOURCE_MANAGER = ResourceManager()


def get_global_manager():
    """Return the global ResourceManager."""
    global RESOURCE_MANAGER
    return RESOURCE_MANAGER


def set_global_manager(manager):
    """Set the global ResourceManager."""
    global RESOURCE_MANAGER
    RESOURCE_MANAGER = manager


@contextlib.contextmanager
def temp_manager(manager):
    """Temporarily change the global resource manager using this with context."""
    old = get_global_manager()
    set_global_manager(manager)
    try:
        yield
    finally:
        set_global_manager(old)


def add_manager(man):
    """Add a manager to help with saving all managers."""
    return get_global_manager().add_manager(man)


def remove_manager(man):
    """Remove a linked manager."""
    return get_global_manager().remove_manager(man)


def clear():
    """Clear out all of the resources that are registered."""
    get_global_manager().clear()


def register(package, name, alias=MISSING, **kwargs):
    """Register a resource. You can optionally have an alias name identifier.
    When using the alias the last resource registered with the same alias will be used.

    Args:
        package (str): Package name ('check_lib.check_sub')
        name (str): Name of the resource ('edit-cut.png')
        alias (str)[MISSING]: Shortcut alias name identifer for the resource.
            ... (Ellipsis) will be the name with the extension (EX: "myimg.png")
            None will be the name without the extension (EX: "myimg")
        **kwargs (dict): Dictionary of keyword arguments to set as attributes to the resource.
    """
    return get_global_manager().register(package, name, alias=alias, **kwargs)


def register_data(data, package, name, alias=MISSING, **kwargs):
    """Register a plain data resource.

    Args:
        data (bytes/str/object)[None]: Plain data to store.
        package (str)[None]: Package name ('check_lib.check_sub')
        name (str)[None]: Name of the resource ('edit-cut.png')
        alias (str)[MISSING]: Shortcut alias name identifer for the resource.
            ... (Ellipsis) will be the name with the extension (EX: "myimg.png")
            None will be the name without the extension (EX: "myimg")
        **kwargs (dict): Dictionary of keyword arguments to set as attributes to the resource.
    """
    return get_global_manager().register_data(data, package, name, alias=alias, **kwargs)


def register_directory(package, directory='', recursive=False, extensions=None, exclude=None, **kwargs):
    """Register all items in a directory.

    Note:
        This does not register the __init__.py module.

    Args:
        package (str): Package name ('check_lib.check_sub')
        directory (str)['']: Additional directory path.
        recursive (bool)[False]: If True iterate through subdirectories and find all files.
        extensions (list/str)[None]: List of extensions to register (".csv", ".txt", "" for "LICENSE" with no ext).
            If None register all.
        exclude (list/str)[None]: List of filenames to exclude.
        **kwargs (dict): Dictionary of keyword arguments to set as attributes to the resource.

    Returns:
        directory (list): List of Resource objects that were registered.
    """
    return get_global_manager().register_directory(package, directory=directory, recursive=recursive,
                                                   extensions=extensions, exclude=exclude, **kwargs)


def register_resource(rsc, **kwargs):
    """Register a resource.

    Args:
        rsc (Resource): Resource object to register and access later.

    Returns:
        rsc (Resource): Resource object that was registered
    """
    return get_global_manager().register_resource(rsc, **kwargs)


def unregister(rsc=None, name=None, alias=None):
    """Unregister a resource.

    Args:
        rsc (str/Resource)[None]: Resource to unregister or package name (with name argument) or alias.
        name (str)[None]: Required if package name given
        alias (str)[None]: Alias name identifier to find. This may also be given as first argument.

    Returns:
        rsc (Resource): Resource that was found

    Raises:
        error (ResourceNotAvailable): Resource was not found!
    """
    return get_global_manager().unregister(rsc=rsc, name=name, alias=alias)


def has_resource(rsc=None, name=None, alias=None):
    """Return if the registered resource exists (this can be the alias, resource, or package_path."""
    return get_global_manager().has_resource(rsc=rsc, name=name, alias=alias)


def get_resources():
    """Return a list of registered resources.

    If multiple resources have the same alias the last one will be used.
    """
    return get_global_manager().get_resources()


def get_resource(rsc, fallback=None, default=MISSING):
    """Return the found Resource object from the given resource, fallback, or default value.

    Args:
        rsc (str/Resource): Alias name identifier, package path, or Resource object that has a registered resource
        fallback (str)[None]: Fallback alias if the given alias was not registered.
        default (object)[MISSING]: Default value to return if the alias and fallback were not found.

    Returns:
        rsc (Resource): The found Resource object.
    """
    return get_global_manager().get_resource(rsc, fallback=fallback, default=default)


def get_binary(rsc, fallback=None, default=MISSING):
    """Return the binary data for the Resource object from the given identifier or default.

    Args:
        rsc (str/Resource): Alias name identifier, package path, or Resource object that has a registered resource
        fallback (str)[None]: Fallback alias if the given alias was not registered.
        default (object)[MISSING]: Default value to return if the alias and fallback were not found.

    Returns:
        data (bytes): Binary data that was read.
    """
    return get_global_manager().get_binary(rsc, fallback=fallback, default=default)


def get_text(rsc, fallback=None, default=MISSING, encoding='utf-8', errors='strict'):
    """Return the text data for the Resource object from the given identifier or default.

    Args:
        rsc (str/Resource): Alias name identifier, package path, or Resource object that has a registered resource
        fallback (str)[None]: Fallback alias if the given alias was not registered.
        default (object)[MISSING]: Default value to return if the alias and fallback were not found.
        encoding (str)['utf-8']: Encoding to convert binary data to text.
        errors (str)['strict']: Error handling code when converting the binary data to text.

    Returns:
        text (str): Text data that was read.
    """
    return get_global_manager().get_text(rsc, fallback=fallback, default=default, encoding=encoding, errors=errors)


def registered_datas(resource_manager=None, use_dest_dirs=True):
    """Return a list of datas that were registered.

    Args:
        resource_manager (ResourceManager)[None]: Resource manger to use. If None use default global ResourceManager.
        use_dest_dirs (bool)[True]: If True the destination will be a directory. If False the dest will be the filename.

    Returns:
        datas (list): List of (existing file path, rel install path).
    """
    if resource_manager is None:
        resource_manager = get_global_manager()

    datas = []

    for man in [resource_manager] + resource_manager.managers:
        for resource in man.get_resources():
            if resource.is_resource():
                with resource.as_file() as rsc_file:
                    package_path = str(resource.package_path)
                    if use_dest_dirs:
                        package_path = os.path.dirname(package_path)
                    data = (os.path.relpath(str(rsc_file)), package_path)
                    if data not in datas:
                        datas.append(data)

    return datas



import os
import types
import contextlib

try:
    from importlib.machinery import SOURCE_SUFFIXES
except (ImportError, Exception):
    SOURCE_SUFFIXES = ['.py', '.pyw', '.py2', '.py3']

from resource_man.__meta__ import version as __version__
from resource_man.importlib_interface import \
    READ_API, FILES_API, Traversable, contents, is_resource, read_binary, read_text, files, as_file
from resource_man.interface import \
    ResourceNotAvailable, Resource, ResourceManagerInterface, ResourceManager, \
    get_global_manager, set_global_manager, temp_manager, add_manager, remove_manager, clear, \
    register_resource, register, register_data, register_directory, unregister, has_resource, \
    get_resources, get_resource, get_binary, get_text, registered_datas, \
    MISSING


__all__ = [
    'find_datas', 'EXCLUDE_EXT', 'SOURCE_SUFFIXES',

    'READ_API', 'FILES_API', 'Traversable', 'contents', 'is_resource', 'read_binary', 'read_text', 'files', 'as_file',

    'ResourceNotAvailable', 'Resource', 'ResourceManagerInterface', 'ResourceManager',
    'get_global_manager', 'set_global_manager', 'temp_manager', 'add_manager', 'remove_manager',
    'clear', 'register_resource', 'register', 'register_data', 'register_directory', 'unregister',
    'has_resource', 'get_resources', 'get_resource', 'get_binary', 'get_text', 'registered_datas',
    'MISSING',
    '__version__'
    ]


EXCLUDE_EXT = SOURCE_SUFFIXES + ['.pyc', '.pyd']


def find_datas(package, exclude_ext=None, use_dest_dirs=True, **kwargs):
    """Collect data files for pyinstaller.

    Args:
        package (types.ModuleType/str/Traversable): Top level package module or module name.
        exclude_ext (list)[None]: List of extensions to not include in the pyinstaller data.
        use_dest_dirs (bool)[True]: If True the destination will be a directory. If False the dest will be the filename.

    Returns:
        datas (list): List of (abs file path, rel install path). This will also include subdirectories.
    """
    if exclude_ext is None:
        exclude_ext = EXCLUDE_EXT

    datas = []
    pkg_name = package
    if isinstance(package, types.ModuleType):
        pkg_name = package.__package__

    with contextlib.suppress(ImportError, Exception):
        toplvl = files(package)
        with as_file(toplvl) as n:
            toplvl_filename = str(n)  # n should be a Path object, but I noticed it was a str anyway
        subdirs = [toplvl]
        while True:
            try:
                directory = subdirs.pop(0)
            except IndexError:
                break

            for path in directory.iterdir():
                if path.name == '__pycache__':
                    continue

                if path.is_dir():
                    # Add subpackage
                    subdirs.append(path)
                elif path.suffix not in exclude_ext:
                    # Assume this is a resource.
                    with as_file(path) as filename:
                        try:
                            filename = str(filename.resolve())
                        except (AttributeError, Exception):
                            pass

                        relpath = os.path.join(pkg_name, os.path.relpath(filename, toplvl_filename))
                        if use_dest_dirs:
                            relpath = os.path.dirname(relpath)

                        data = (filename, relpath)
                        if data not in datas:
                            datas.append(data)

    return datas
